/*
 * pluginconsole.c
 * A simple console to set the output values of a dmx-plugin.
 *
 * Copyright (C) 1999,2000,2001  Michael Stickel
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include <stdio.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <unistd.h>
#include <stdlib.h>
#include <fcntl.h>
#include <signal.h>
#include <curses.h>

#include <dmx/plugin.h>
#include <trycatch.h>

typedef unsigned long Zeit;

#define KEY_ESCAPE      (27)

#if 0
#define KAN_PER_LINE    (25)
#define KAN_OUT	        (KAN_PER_LINE * 12)
#define MAX_KAN_OUT	(KAN_OUT)
#else

int columns = 25;
int rows    = 12;
#define KAN_PER_LINE    (columns>50?50:columns)
#define KAN_OUT	        (KAN_PER_LINE * (rows>30?30:rows))
#define MAX_KAN_OUT	(50*30)
#endif


#define OUTP_Y          (0)

#define FL              (100)

static int   chan = 0;
static int   output[MAX_KAN_OUT] = {0};
static int   gmaster=100;
static char  blackout = 0;

static  dmxplugin_t *plugin = NULL;

static void put_mask ();

static void init_mask ()
{
  int  i;

  refresh ();
  
  attron (COLOR_PAIR(1));
  mvprintw (OUTP_Y,0, "                                 DMX CONSOLE     (q=Quit)             ");
  attroff (COLOR_PAIR(1));
  
  attron (COLOR_PAIR(5));
  for (i=0; i<KAN_OUT; i++)
    {
      if (i%KAN_PER_LINE==0)
	move (OUTP_Y+1+(i/KAN_PER_LINE)*2, 1);

      if ( (i%KAN_PER_LINE==0) || (i%KAN_PER_LINE==(KAN_PER_LINE-1)) )
	printw ("%03d ", i+1);
      else
	printw ("%02d ", (i+1)%100);

    }
  attroff (COLOR_PAIR(5));

  put_mask ();
}


static void put_mask ()
{
  int i;

  if (plugin)
    {
      dmxuniverse_t universe;

      universe.numslots = (MAX_KAN_OUT>512)?512:MAX_KAN_OUT;
      memset(universe.buffer, 0, 512);

      for (i=0; i<KAN_OUT; i++)
        {
          int val = 0;
          if (!blackout)
            {
	      val = ((float)output[i] * 2.56 * (float)(gmaster/100.0));
              if (val>255) val=255;
              if (val<0) val=0;
            }
	  universe.buffer[i] = (unsigned char)val;
        }

      if (!plugin->send_universe || plugin->send_universe(0, &universe) < 0)
        mvprintw(0,60,"xmit-error");
      refresh ();
    }

  attron (COLOR_PAIR(1));
  mvprintw (OUTP_Y,1, "channel: %-3d value: %-3d", chan+1, output[chan]);
  if (blackout)
    mvprintw (OUTP_Y,80-10, "BLACKOUT");
  else
    mvprintw (OUTP_Y,80-10, (gmaster<100)?"   GM=%02d":"   GM=FL", gmaster);
  attroff (COLOR_PAIR(1));
  
  
  for (i=0; i<KAN_OUT; i++)
    {
      if (i==chan)
	attron (A_REVERSE);
      switch (output[i])
	{
	case 0:   attron (COLOR_PAIR(4)); break;
	case 100: attron (COLOR_PAIR(3)); break;
	default:  attron (COLOR_PAIR(2)); break;
	}
      if (i%KAN_PER_LINE==0)
	move (OUTP_Y+2+(i/KAN_PER_LINE)*2, 2);
      printw ((output[i]>99)?"FL":(output[i]>0)?"%02d":"__", output[i]);
      
      switch (output[i])
	{
	case 0:   attroff (COLOR_PAIR(4)); break;
	case 100: attroff (COLOR_PAIR(3)); break;
	default:  attroff (COLOR_PAIR(2)); break;
	}
      attroff (A_REVERSE);
      printw (" ");
    }
  refresh ();
}


void issue ()
{
  printf ("pluginconsole [-? | -c columns | -r rows | -p <plugin>]\n");
}

int main (int argc, char **argv)
{
  char    *pluginname = "avabudp";
  WINDOW  *w = NULL;

  if (argc>1)
    {
      int i=1;
      for (i=1; i<argc; i++)
        {
          if (*argv[i] == '-')
            {
              switch (argv[i][1])
                {
                  case 'p':
                    if (i+1<argc)
                      pluginname = argv[++i];
                    break;

                  case 'c':
                    if (i+1<argc)
                      columns = atoi(argv[++i]);
                    break;

                  case 'r':
                    if (i+1<argc)
                      rows = atoi(argv[++i]);
                    break;

                  case '?':
                    issue ();
                    exit(0);
                    break;
                }
            }
        }
    }


  w = initscr ();
  if (!w)
    {
      printf ("unable to initialize output\n");
      exit(0);
    }

  plugin = loadPlugin (getenv("DMXPLUGINPATH"), pluginname);
  if (!plugin)
    {
      endwin();
      fprintf(stderr, "unable to load output-plugin %s\n", pluginname);
      exit(0);
    }
 

  start_color ();
  savetty ();
  
  init_pair (0, COLOR_BLACK,  COLOR_WHITE);
  init_pair (1, COLOR_WHITE,  COLOR_BLUE);
  init_pair (2, COLOR_GREEN,  COLOR_BLACK);
  init_pair (3, COLOR_RED,    COLOR_BLACK);
  init_pair (4, COLOR_WHITE,  COLOR_BLACK);
  init_pair (5, COLOR_BLACK,  COLOR_YELLOW /*WHITE*/);
  
  
  if (w)
    {
      int  state;
      int  ende=0;
      int  bptr=0;
      
      noecho ();
      raw ();
      keypad (w, TRUE);
      
      init_mask ();
      
      state = 0;
      bptr = 0;
      while (!ende)
	{
	  int c = wgetch (w);
	  switch (c)
	    {
	    case 'q':
	    case KEY_ESCAPE:
	      ende++;
	      break;

	    case KEY_F(9): 
	      blackout = blackout?0:1;
	      put_mask ();
	      break;

	    case KEY_F(7):
	      gmaster = (--gmaster < 0)?0:gmaster;
	      put_mask ();
	      break;

	    case KEY_F(8):
	      gmaster = (++gmaster > 100)?100:gmaster;
	      put_mask ();
	      break;
	      
	      
	      
	      
	    case KEY_PPAGE:
	      output[chan]+=9;
	    case '+':
	      output[chan] = (output[chan]+1 > 100)?100:output[chan]+1;
	      put_mask ();
	      break;
	      
	    case KEY_NPAGE:
	      output[chan]-=9;
	    case '-':
	      output[chan] = (output[chan]-1 < 0)?0:output[chan]-1;
	      put_mask ();
	      break;
	      

	    case ' ':
	      output[chan] = (output[chan]>0) ? 0 : 100;
	      put_mask ();
	      break;

	    case '0':
	      output[chan] = 100;
	      put_mask ();
	      break;

	    case '1':
	    case '2':
	    case '3':
	    case '4':
	    case '5':
	    case '6':
	    case '7':
	    case '8':
	    case '9':
	      output[chan] = (c-'0')*10;
	      put_mask ();
	      break;


	    case KEY_RIGHT:
	      chan++;
	      chan %= KAN_OUT;
	      put_mask ();
	      break;

	    case KEY_LEFT:
	      chan = (chan-1) + KAN_OUT;
	      chan %= KAN_OUT;
	      put_mask ();
	      break;

	    case KEY_DOWN:
	      chan = chan+KAN_PER_LINE;
	      chan %= KAN_OUT;
	      put_mask ();
	      break;

	    case KEY_UP:
	      chan = (chan-KAN_PER_LINE) + KAN_OUT;
	      chan %= KAN_OUT;
	      put_mask ();
	      break;

	      
	    default:
	      mvprintw (1,1,"%03X", c);
	      refresh ();
	      break;
	    }
	}
    }
  
  noraw ();
  echo ();
  
  resetty (); 
  endwin ();

  if (plugin && plugin->delete)
    plugin->delete (plugin);
  plugin=NULL;

  printf ("\nexited Light-Console\n");
  return 0;
}




