/*
 * Implementation of a plugin module to
 * access the DMX over Ethernet protocol.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>

#include <unistd.h>
#include <fcntl.h>
#include <pthread.h>
#include <signal.h>

#include <errno.h>

#include <netdb.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include <sys/ioctl.h>
#include <linux/if.h>

#include <dmx/plugin.h>

#include "dmxoe.h"


#undef DEBUG

#ifdef DEBUG
# define DEBUG_FUNC(x) x
#else
# define DEBUG_FUNC(x)
#endif


/*
 * ---------[ Prototypes ]----------
 */
static int Initialize (char * hostname);

static void *DMXoE_Thread (void *arg);

static int isMyself (struct sockaddr_in *addr);
static void add_interfaces2nonhosts();

static size_t shownet_encode (U8 dmxbuf[], size_t dmxlen, U8 netbuf[], size_t netlen);
static size_t shownet_decode (U8 pktbuf[], size_t pktlen, U8 netbuf[], size_t netlen);

static int    sendDMX         (int universe, unsigned char *dmxbuffer, int dmxlen,
                               struct sockaddr *sa, socklen_t len);

static int    activateReceiver ();
static int    stopReceiver     ();

static char * IP2str          (char *str, struct sockaddr_in *addr);




/*
 * -------------[ plugin wide variables ]-------------
 */

static  dmxuniverse_t       *universes[10];
static  const int max_universes =  sizeof(universes) / sizeof(dmxuniverse_t *);
static  int  used_universes = 0;

static  pthread_t rxtx_thread;
static  char      thread_active = 0;


static  int subnet = 15;
static  int debug  = 0;
static  int sock   = -1;


static  struct sockaddr_in servaddr;
static  struct sockaddr_in anyaddr;
static  struct sockaddr_in broadcast;


/*
 * list of socketaddresses that should be ignored
 */
static  struct sockaddr_in  nohosts[10]; // buggy constant, but will work now.
static  int    num_nohosts = 0;



/*
 * -------------[ plugin methods ]-------------
 */

/*
 * This method is called by the plugin interface after the
 * module has been loaded.
 */
int initialize (void)
{
  int i;

  for (i=0; i<max_universes; i++)
    universes[i] = NULL;

  if (!Initialize (NULL))
    return 0;
  return -1;
}

/*
 * This method is called by the plugin interface before the
 * module will be unloaded.
 */
int finalize ()
{
  int i;

  if (thread_active)
    {
      stopReceiver ();
    }

  for (i=0; i<max_universes; i++)
    if (universes[i])
      universes[i] = NULL;

  if (sock != -1)
    {
      shutdown (sock, 3);
      close (sock);
    }
}


/*
 * This method binds a universe object to
 * the universe number <nr>.
 * Returns 0 on success or -1 on error.
 */
int bind_universe (int nr, dmxuniverse_t *u)
{
  if (nr>=0 && nr<max_universes && !universes[nr])
    {
      if (used_universes==0)
        {
          activateReceiver();
        }
      used_universes++;

      universes[nr] = u;
      return 0;
    }
  return -1;
}


/*
 * This method tries release the binding between
 * a universe object and the universe number <nr>.
 * Returns 0 on success or -1 on error.
 */
int unbind_universe (int nr, dmxuniverse_t *u)
{
  if (nr>=0 && nr<max_universes && universes[nr] && universes[nr]==u)
    {
      universes[nr] = NULL;
      used_universes--;
      if (used_universes<=0)
        {
          /* should stop the thread */
        }
      return 0;
    }
  return -1;
}


/*
 * Sends out the universe data of the universe
 * object <u> on the universe <nr>.
 */
int send_universe (int nr, dmxuniverse_t * u)
{
  if (u && nr>=0 && nr<max_universes /* && universes[nr] && universes[nr]==u */)
    {
      return sendDMX (nr, u->buffer, u->numslots, (struct sockaddr *)&broadcast, sizeof(broadcast));
    }
  return -1;
}


/*
 * Get the some information about the universe.
 */
int get_plugin_info (dmxplugin_info_t *info)
{
  if (info)
    {
      strcpy (info->vendor_name, "Strand Lighting");
      strcpy (info->proto_name, "DMX over Ethernet");
      info->proto_release = 0.0;

      info->release = 1.0;
      strcpy (info->author, "Michael Stickel");
      strcpy (info->email,  "michael@cubic.org");
      info->num_universes = max_universes;
      return 0;
    }
  return -1;
}




static int getInetAddrByNameAndPort (struct sockaddr_in *sa, size_t slen, unsigned short port, char *name)
{
  if (sa && slen==sizeof(struct sockaddr_in))
    {
      struct hostent * he = gethostbyname(name);
      if (he)
        {
          bzero (sa, slen);
          sa->sin_family = AF_INET;
          sa->sin_port = port;
          memcpy (&sa->sin_addr.s_addr, he->h_addr, he->h_length);
          return 0;
        }
    }
  return -1;
}


static int Initialize (char * hostname)
{
  int  one    = 1;
  int *oneptr = &one;


  sock = socket (AF_INET, SOCK_DGRAM, 0);
  if (sock<0)
    {
      printf ("Error createing socket: %s\n", strerror(errno));
      return -1;
    }
  
  if (setsockopt (sock, SOL_SOCKET, SO_BROADCAST, oneptr, sizeof(one)) == -1)
    {
      fprintf (stderr, "Error getting broadcast permission: %s\n", strerror(errno));
    }

  getInetAddrByNameAndPort (&servaddr,  sizeof(servaddr),  htons(DMXOE_DEFAULT_PORT), hostname?hostname:"255.255.255.255");
  getInetAddrByNameAndPort (&anyaddr,   sizeof(anyaddr),   htons(DMXOE_DEFAULT_PORT), "0.0.0.0");
  getInetAddrByNameAndPort (&broadcast, sizeof(broadcast), htons(DMXOE_DEFAULT_PORT), "255.255.255.255");

  memcpy (&nohosts[num_nohosts++], &anyaddr, sizeof(anyaddr));
  memcpy (&nohosts[num_nohosts++], &broadcast, sizeof(broadcast));

  add_interfaces2nonhosts ();

  return 0;
}

/*
 * add the addresses of all network interfaces into the list of ip's to ignore
 */
static void add_interfaces2nonhosts()
{
  struct ifconf ifc;
  struct ifreq  ifr[50];

  int sockfd = socket(AF_INET, SOCK_DGRAM,0);
  if (sockfd!=-1)
    {
      ifc.ifc_len = sizeof(ifr);
      ifc.ifc_req = ifr;
      if (ioctl (sockfd, SIOCGIFCONF, &ifc) >= 0)
        {
          int count = ifc.ifc_len / sizeof(struct ifreq);
          int i;

          for (i=0; i<count; i++)
            {
              memcpy (&nohosts[num_nohosts], &ifr[i].ifr_addr, sizeof(struct sockaddr_in));
              nohosts[num_nohosts].sin_port = htons(DMXOE_DEFAULT_PORT);
              nohosts[num_nohosts].sin_family = AF_INET;

              num_nohosts++;
            }
        }
      close(sockfd);
    }
}






static int isMyself (struct sockaddr_in *addr)
{
  if (addr)
    {
      int i;
      for (i=0; i<num_nohosts; i++)
        if (addr->sin_addr.s_addr==nohosts[i].sin_addr.s_addr && addr->sin_port==nohosts[i].sin_port)
          return 1;
      return 0;
    }
  return -1;
}





char *IP2str (char *str, struct sockaddr_in *addr)
{
  if (str)
    {
      if (addr)
	{
	  sprintf (str, "%s:%d", inet_ntoa (addr->sin_addr), ntohs(addr->sin_port));
	}
      else
	strcpy (str, "unknown");
    }
  return str;
}



static int activateReceiver ()
{
  if (!thread_active)
    {
      int stat=0;

      if (bind (sock, (struct sockaddr *)&anyaddr, sizeof(anyaddr)) < 0)
        {
          int err = errno;
          char * str = "unknown";
          if (err==EINVAL)
            str = "EINVAL";
          else if (err==EACCES)
            str = "EACCES";
          printf ("Error while bind %s\n", str);
          return -1;
        }

      stat = pthread_create (&rxtx_thread, NULL, DMXoE_Thread, NULL);
      if (stat)
	{
	  thread_active = 0;
	  printf ("unable to create thread\n");
	}
      else
	{
	  thread_active = 1;
	  return 0;
	}
    }
  return -1;
}


static int stopReceiver ()
{
  if (thread_active)
    {
      int stat = pthread_kill (rxtx_thread, SIGINT);
      if (!stat || stat == ESRCH)
	{
	  void *ret = NULL;
	  if (!pthread_join (rxtx_thread, &ret))
	    {
	      printf ("thread returned successfully and returned value %p\n", ret);
              thread_active = 0;
	      return 0;
	    }
	}
    }
  return -1;
}




static void *DMXoE_Thread (void *arg)
{
  int               dglen;
  unsigned char     buffer[2000];

  /*
   * we have to setup the sceduleing policy to fifo to have
   * a quicker response.
   */

  while (1)
    {
      struct sockaddr_in claddr;
      socklen_t          cllen = sizeof(claddr);

      dglen = recvfrom (sock, buffer, sizeof(buffer), 0, (struct sockaddr *)&claddr, &cllen);
      if (dglen < 0)
	{
          int e = errno;
	  printf ("Error on receive:%d:%d: %s\n", dglen, e, strerror(e));
	}
      else if (dglen>=sizeof(header) && !isMyself (&claddr))
        {
          header *phead = (header *)buffer;
          char dmxbuffer[4096];

          if (phead->version == DMX_OVER_ETHERNET_VERSION &&
              phead->long_sig0[0] == 0 &&
              phead->long_sig0[1] == 0 &&
              phead->long_sig0[2] == 0x58 &&
              phead->long_sig0[3] == 2)
            {
              size_t dmxlen;

              memset (dmxbuffer, 0, sizeof(dmxbuffer));
              dmxlen = shownet_decode (buffer, dglen, dmxbuffer, sizeof(dmxbuffer));
              if (dmxlen > 0)
                {
                  int i;
                  for (i=0; i < ((dmxlen-1)/512)+1; i++)
                    {
                      dmxuniverse_t *u = universes[i];
                      if (u && u->notify)
                        {
                          u->numslots = (dmxlen-i*512 > 512)?512:(dmxlen-i*512);
                          memcpy (u->buffer, &dmxbuffer[i*512], u->numslots);
                          u->notify (u);
                        }
                    }
                }
            }
        }
    }
}

int sendDMX (int nr, unsigned char *dmxbuffer, int dmxlen, struct sockaddr *tsa, socklen_t len)
{
  if (dmxbuffer && dmxlen>0 && tsa && len>0)
    {
      U8 netbuf[2000];
      size_t pktsize;

      pktsize = shownet_encode (dmxbuffer, dmxlen, netbuf, sizeof(netbuf));
      if (pktsize>0)
        {
          struct sockaddr_in  sa;

          memcpy (&sa, tsa, len); /* copy template to real socket address buffer */
          sa.sin_port = htons(DMXOE_DEFAULT_PORT);
          sa.sin_family = AF_INET;

          if (sendto (sock, netbuf, pktsize, 0, (struct sockaddr *)&sa, sizeof(sa)) < 0)
            {
              printf ("avabudp.sendDMX : Error sending: %s\n", strerror(errno));
              printf ("sock=%d\n", sock);
            }
        }
    }
}


/*---------------------------------------------------------------------------------------------------------
 * dmxoe_codec.c
 */


static  U16 packetSequence = 1;
static  U32 my_ip = (192<<24) | (168<<16) | (0<<8) | 49;

static  U16 getPacketSequence()
{
  return packetSequence++;
}


/*
 *   1-100 slots     one block contains all the slots
 * 101-200 slots     two blocks: the slots are split evenly between
 *                   them with the first block getting the extra.
 *                   eg. for 121 slots the first block would contain
 *                   the first 61 slots, and the second block would
 *                   contain the last 60 slots.
 *     201 slots     first block has 51 slots, second block has 50 slots,
 *                   third block has 100 slots
 *     202 slots     first block has 51 slots, second bock has 50 slots,
 *                   third block has 50 slots, fourh block has 51 slots.
 *     203 slots     51, 50, 51, 51
 *     204 slots     51, 51, 51, 51
 *     205 slots     52, 51, 51, 51
 *     206 slots     52, 51, 51, 52
 *     207 slots     52, 51, 52, 52
 *     208 slots     52, 52, 52, 52
 *     209 slots     53, 52, 52, 52
 *     210 slots     53, 52, 52, 53
 *     and so on for the rest.
 */
static long getSlotLayout (int count, U16 *lengths)
{
  U16 c[4];

  memset (c, 0, sizeof(c));

  if (count>=1 && count<=100)
    {
      DEBUG_FUNC(printf ("case1\n"));
      c[0] = count;
    }
  else if (count>=101 && count<=200)
    {
      DEBUG_FUNC(printf ("case2\n"));
      c[1] = count/2;
      c[0] = count - c[1];
    }
  else if (count==201)
    {
      DEBUG_FUNC(printf ("case3\n"));
      c[0] = 51;
      c[1] = 50;
      c[2] = 100;
    }
  else if (count >= 202)
    {
      register int t1 = count/4;
      register int t2 = count%4;
      c[0] = t1 + ((t2>0)?1:0);
      c[3] = t1 + ((t2>1)?1:0);
      c[2] = t1 + ((t2>2)?1:0);
      c[1] = t1;
      DEBUG_FUNC(printf ("case4\n"));
    }

  if (lengths)
    memcpy (lengths, c, sizeof(U16)*4);

  return (c[3]<<24) | (c[2]<<16) | (c[1]<<8) | c[0];
}


/*
 *
 *
 */
static size_t rle_encode (U8 *dstbuf, size_t dstlen, U8 *srcbuf, size_t srclen)
{
  size_t cnt = 0;
  size_t len = srclen;

  /* dstlen = ((srclen / 0x7f) * (0x7f+1)) */
  if (srclen + (srclen/0x7f) > dstlen)
    {
      DEBUG_FUNC(printf ("rle_encode: error, target to small src=%i dst=%i\n", srclen, dstlen));
      return -1;
    }

  while (len>0)
    {
      int d = (len > 0x7f) ? 0x7f : len;

      *(dstbuf++) = d;
      memcpy (dstbuf, srcbuf, d);
      dstbuf += d;
      srcbuf += d;
      cnt += d;
      len -= d;
    }
  return cnt;
}




static size_t rle_decode (U8 srcbuf[], size_t srclen, U8 dstbuf[], size_t dstlen)
{
  U8  count = 0;
  int src   = 0;
  int dst   = 0;

  if(srclen <= 2 && (srcbuf[0] & 0x80) == 0)
    while(src<srclen && dst<dstlen)
      dstbuf[dst++] = srcbuf[src++];
  else
    while(src<srclen && dst<dstlen)
      {
        count = srcbuf[src++];
        if(count & 0x80)
          {
            count &= ~0x80;
            memset(&dstbuf[dst], srcbuf[src], count);
            dst += count;
            src++;
          }
        else
          {
            memcpy(&dstbuf[dst], &srcbuf[src], count);
            dst += count;
            src += count;
          }
      }
  return dst;
}



static size_t shownet_encode (U8 dmxbuf[], size_t dmxlen, U8 netbuf[], size_t netlen)
{
  header *phead = (header *)netbuf;
  int     rle_offset=0;
  int     i;

  if (sizeof(header) != HEADER_SIZE)
    {
      DEBUG_FUNC(printf("FATAL: illegal size of paket-structure\n"));
      return -1;
    }

  if (dmxlen <= 0)
    return 0;

  phead->version  = DMX_OVER_ETHERNET_VERSION;
  phead->srcip    = htonl(my_ip);
  phead->sequence = htons(getPacketSequence());
  phead->long_sig0[0] = 0;
  phead->long_sig0[1] = 0;
  phead->long_sig0[2] = 0x58; /* 0x258 = 600 (zufall?) */
  phead->long_sig0[3] = 2;
  memcpy(phead->devicename, "console1", 8);
  phead->long_sig1[0] = 0;
  //  rle encoded dimmer data

  getSlotLayout (dmxlen, phead->length);

  for (i = 0; i<4; i++)
    {
  /*    phead->length[i] = rle_length; */

      phead->slot[i] = 1 + rle_offset;
      phead->rle_offset[i] = RLE_OFFSET_BIAS + rle_offset;
      phead->rle_offset[i+1] = phead->rle_offset[i] + phead->length[i];

      rle_encode (phead->data + rle_offset,
                  (netlen-sizeof(header)) / 4,
                  dmxbuf + phead->slot[i] - 1,
                  phead->length[i]
                 );

      rle_offset += phead->length[i];
    }
  return HEADER_SIZE + rle_offset;
}




static size_t shownet_decode (U8 pktbuf[], size_t pktlen, U8 netbuf[], size_t netlen)
{
  header *phead = (header*) pktbuf;
  int     rle_offset,
          rle_length;
  int     i;
  size_t  len = 0;

  for (i = 0; i<4; i++)
    {
      size_t t=0;
      if(phead->slot[i] == 0) break;
      rle_offset = phead->rle_offset[i] - RLE_OFFSET_BIAS;
      rle_length = phead->rle_offset[i+1] - phead->rle_offset[i];
      t = rle_decode (phead->data + rle_offset,
                  rle_length,
                  netbuf + phead->slot[i] - 1,
                  -1
                 );
      if (t>0 &&  phead->slot[i]-1+t> len)
        len = phead->slot[i]-1+t;
    }
  return len;
}

